#include<Servo.h>

#define intCallback 0
#define TrigPin 3
#define PanPin 4
#define TiltPin 5
#define FirePin 6
#define EchoPin 7
#define LEDPin 13

#define WaitMin 1000
#define WaitMax 1000

#define PanMin 50
#define PanMax 130
#define TiltMin 8
#define TiltMax 50

#define MinRange 20
#define MaxRange 80

enum DCState {
  stMove, stAim, stFireAtWill, stReload};

enum DCMode {
  mdRandom, mdScan, mdStatic};

struct Vector 
{
  int x;
  int y;
};

Servo servoPan;
Servo servoTilt;

int wantedPositionPan = 90;
int wantedPositionTilt = TiltMin;
int currentPositionPan = wantedPositionPan;
int currentPositionTilt = wantedPositionTilt;

int currentDirectionPan;
int currentDirectionTilt;

int servoStep = 1;
int servoStepDelay = 15;
int servoStepElapsed;
long lastMillis;

long waitTime;
long waitTimeElapsed;

int shootingDelay = 6 * 150;

int shotsFired = 0;
bool fired = false;

DCState state = stAim;

//DCMode mode = mdRandom;
DCMode mode = mdScan;
//DCMode mode = mdStatic;

int currentVector = -1;

Vector vectors[9] = {
  { 
    50, 8                      }
  ,
  { 
    50, 29                      }
  ,
  {
    50, 50                      }
  ,
  {
    90, 8                      }
  ,
  {
    90, 29                      }
  ,
  {
    90, 50                      }
  ,
  {
    130, 8                      }
  ,
  {
    130, 29                      }
  ,
  {
    130, 50                      }
};

void setup()
{ 
  Serial.begin(9600);

  randomSeed(analogRead(0));

  pinMode(TrigPin, OUTPUT);
  pinMode(EchoPin, INPUT);

  pinMode(FirePin, OUTPUT);
  digitalWrite(FirePin, LOW);

  servoPan.attach(PanPin);
  servoTilt.attach(TiltPin);  

  servoPan.write(currentPositionPan);
  servoTilt.write(currentPositionTilt);

  pinMode(LEDPin, OUTPUT);
  digitalWrite(LEDPin, HIGH);

  attachInterrupt(intCallback, MissileLaunched, RISING);

  delay(5000);

  digitalWrite(LEDPin, LOW);

  lastMillis = millis();
}

void loop()
{ 
  switch(state)
  {
  case stAim:
    setWantedPosition();
    break;
  case stMove:
    move();
    break;
  case stFireAtWill:
    fireAtWill();
    break;
  case stReload:
    servoPan.detach();
    servoTilt.detach();
    break;
  }
}

void setWantedPosition()
{
  switch(mode)
  {
  case mdRandom:
    setRandomWantedPosition();
    break;
  case mdScan:
    setScanWantedPosition();
    break;
  case mdStatic:
    setReloadPosition();
    break;
  }

  fired = false;
  state = stMove;
  digitalWrite(LEDPin, LOW);
}

void setRandomWantedPosition()
{
  while(wantedPositionPan == currentPositionPan || wantedPositionTilt == currentPositionTilt)
  {
    wantedPositionPan = random(PanMin, PanMax);
    wantedPositionTilt = random(TiltMin, TiltMax);
  }

  currentDirectionPan = wantedPositionPan < currentPositionPan ? -1 : 1;
  if (wantedPositionPan == currentPositionPan)
  {
    currentDirectionPan = 0;
  }

  currentDirectionTilt = wantedPositionTilt < currentPositionTilt ? -1 : 1;
  if (wantedPositionTilt == currentPositionTilt)
  {
    currentDirectionTilt = 0;
  }  
}

void setScanWantedPosition()
{
  currentVector++;

  currentVector %= 9;

  wantedPositionPan = vectors[currentVector].x;
  wantedPositionTilt = vectors[currentVector].y;

  currentDirectionPan = wantedPositionPan < currentPositionPan ? -1 : 1;
  if (wantedPositionPan == currentPositionPan)
  {
    currentDirectionPan = 0;
  }

  currentDirectionTilt = wantedPositionTilt < currentPositionTilt ? -1 : 1;
  if (wantedPositionTilt == currentPositionTilt)
  {
    currentDirectionTilt = 0;
  }
}

void setReloadPosition()
{
  wantedPositionPan = 90;
  wantedPositionTilt = 0;

  currentDirectionPan = wantedPositionPan < currentPositionPan ? -1 : 1;
  if (wantedPositionPan == currentPositionPan)
  {
    currentDirectionPan = 0;
  }

  currentDirectionTilt = wantedPositionTilt < currentPositionTilt ? -1 : 1;
  if (wantedPositionTilt == currentPositionTilt)
  {
    currentDirectionTilt = 0;
  }  
}

void move()
{
  long currentMillis = millis();                          
  servoStepElapsed += currentMillis - lastMillis;         
  lastMillis = currentMillis;

  if (servoStepElapsed > servoStepDelay)
  {
    servoStepElapsed = 0;
    currentPositionPan += servoStep * currentDirectionPan;
    currentPositionTilt += servoStep * currentDirectionTilt;

    servoPan.write(currentPositionPan);
    servoTilt.write(currentPositionTilt);

    if (currentPositionPan == wantedPositionPan)
    {
      currentDirectionPan = 0;
    }

    if (currentPositionTilt == wantedPositionTilt)
    {
      currentDirectionTilt = 0;
    }

    if (currentPositionPan == wantedPositionPan && currentPositionTilt == wantedPositionTilt)
    {       
      state = stFireAtWill;
      waitTime = random(WaitMin, WaitMax);
      waitTimeElapsed = 0;
      
      if (shotsFired == 6)
      {
        state = stReload;
      }
    }
  }
}

void fireAtWill()
{    
  if (fired == false)
  {
    // 10 ms delay. For some reason it finds matching readings when pinging immediately after the servos are done (???)
    delay(10);

    // Send three pings. If any two are matching, fire.    
    int firstReading = getDistance();
    int secondReading = getDistance();
    int thirdReading = getDistance();

    if ((firstReading == secondReading) && between(firstReading, MinRange, MaxRange) || 
      (secondReading == thirdReading) && between(secondReading, MinRange, MaxRange) || 
      (firstReading == thirdReading) && between(firstReading, MinRange, MaxRange))
    {
      shoot();
    }
  }

  long currentMillis = millis();
  waitTimeElapsed += currentMillis - lastMillis;
  lastMillis = currentMillis;

  if (waitTimeElapsed > waitTime)
  {
    state = stAim;
  }
}

int getDistance()
{
  int distance;
  unsigned long pulseDuration = 0;

  digitalWrite(TrigPin, LOW);
  delayMicroseconds(5);

  digitalWrite(TrigPin, HIGH);
  delayMicroseconds(5);
  digitalWrite(TrigPin, LOW);

  pulseDuration = pulseIn(EchoPin, HIGH);
  pulseDuration = pulseDuration / 2;

  distance = int(pulseDuration / 29);

  return distance;
}

bool between(int value, int min, int max)
{
  return min <= value && max >= value;
}

void shoot()
{
  digitalWrite(FirePin, HIGH);
  shotsFired++;
  fired = true;
}

void MissileLaunched()
{
  digitalWrite(FirePin, LOW);
  digitalWrite(LEDPin, HIGH);

  if (shotsFired == 6)
  {
    mode = mdStatic;
    state = stMove;
    setReloadPosition();
    
    digitalWrite(LEDPin, LOW);
  }
}












